const char *mainid = "$Id: markUniqueUnique.c,v 1.10 2010/12/13 03:40:56 brianwalenz Exp $";

//  Assembly terminator module. It is the backend of the assembly
//  pipeline and replaces internal accession numbers by external
//  accession numbers.

#include  <stdio.h>
#include  <stdlib.h>
#include  <string.h>
#include  <unistd.h>
#include  <assert.h>

#include "AS_global.h"
#include "AS_UTL_UID.h"
#include "AS_UTL_Hash.h"
#include "AS_UTL_Var.h"
#include "AS_MSG_pmesg.h"
#include "MultiAlignStore.h"


#define DEFAULT_UNITIG_LENGTH 1000
#define MAX_LINE_LENGTH 256
 
VA_DEF(uint32)

int main (int argc, char *argv[]) {
   char    *asmFileName    = NULL;
   char    *tigStoreName   = NULL;
   char    *utgFileName    = NULL;
   uint32   tigStoreVers   = 2;

   int      minLength      = DEFAULT_UNITIG_LENGTH;

   uint32   numToggled     = 0;

   argc = AS_configure(argc, argv);
  
   int arg=1;
   int err=0;
   while (arg < argc) {
      if        (strcmp(argv[arg], "-a") == 0) {
         asmFileName = argv[++arg];

      } else if (strcmp(argv[arg], "-t") == 0) {
        tigStoreName = argv[++arg];
        tigStoreVers = atoi(argv[++arg]);

      } else if (strcmp(argv[arg], "-l") == 0) {
         minLength = atoi(argv[++arg]);
      
      } else if (strcmp(argv[arg], "-f") == 0) {
         utgFileName = argv[++arg];

      } else {
         fprintf(stderr, "%s: unknown option '%s'\n", argv[0], argv[arg]);
         err++;
      }

      arg++;
   }

   if (minLength <= 0) err++;

   if ((asmFileName == NULL) || (tigStoreName == NULL) || (utgFileName == NULL) || (err > 0)) {
      fprintf(stderr, "usage: %s -a asmFile -t tigStore version -f utgToMarkUniqueList [-l minLength] \n", argv[0]);
      fprintf(stderr, "\n");
      fprintf(stderr, "  -a asmFile            path to the assembly .asm file\n");
      fprintf(stderr, "  -t tigStore version   path to the tigStore and version to modify\n");
      fprintf(stderr, "  -f utgFile            path the the file listing one UID per line for unitigs to toggle\n");
      fprintf(stderr, "  -l minLength          minimum size of a unitig to be toggled, default=%d)\n", DEFAULT_UNITIG_LENGTH);
      fprintf(stderr, "\n");
      fprintf(stderr, "  Labels surrogate unitigs as non-repeat if the UID is listed in file utgFile\n");
      exit(1);
   }
  
   HashTable_AS      *UIDtoIID         = CreateScalarHashTable_AS();
   HashTable_AS      *UIDtoToggle      = CreateScalarHashTable_AS();
   VA_TYPE(int32)    *unitigLength      = CreateVA_int32(8192);
   VA_TYPE(uint32)   *toggleCount      = CreateVA_uint32(8192);
   
   GenericMesg    *pmesg;
   FILE           *infp = fopen(utgFileName, "r");
   while (!feof(infp)) {
      char uidString[MAX_LINE_LENGTH];
      fgets(uidString, MAX_LINE_LENGTH, infp);
      InsertInHashTable_AS(UIDtoToggle, AS_UID_toInteger(AS_UID_lookup(uidString, NULL)), 0, (uint64)1, 0);
   }
   fclose(infp);

   infp = fopen(asmFileName, "r");   
   while ((EOF != ReadProtoMesg_AS(infp, &pmesg))) {
      SnapUnitigMesg    *utg     = NULL;
      SnapConConMesg    *ctg     = NULL;
      SnapScaffoldMesg  *scf     = NULL;
      uint32             count   = 0;
      uint32             forward = TRUE;
      uint32             lastCtg = 0;

      switch(pmesg->t) {
         case MESG_UTG:
            utg = (SnapUnitigMesg*)(pmesg->m);
            Setint32(unitigLength, utg->iaccession, &utg->length);

            if (utg->length >= minLength && (utg->status == AS_NOTREZ || utg->status == AS_SEP)) {               
               // store the mapping for this unitig's UID to IID and initialize it's instance counter at 0
               if (ExistsInHashTable_AS(UIDtoToggle, AS_UID_toInteger(utg->eaccession), 0)) {
                  count = 1;
                  InsertInHashTable_AS(UIDtoIID, AS_UID_toInteger(utg->eaccession), 0, (uint64)utg->iaccession, 0);               
                  Setuint32(toggleCount, utg->iaccession, &count);
               }
            }
            break;    

         default:
            break;
      }
   }
   fclose(infp);
  
   // open the tig store for in-place writing (we don't increment the version since CGW always reads a fixed version initially)
   // this also removes any partitioning

   MultiAlignStore *tigStore = new MultiAlignStore(tigStoreName, tigStoreVers, 0, 0, TRUE, TRUE);

   for (uint32 i = 0; i < tigStore->numUnitigs(); i++) {
      uint32 *length   = Getuint32(unitigLength, i);
      uint32 *toggle   = Getuint32(toggleCount, i);

      bool toggled = false;
      
      if (toggle != NULL && (*toggle) == 1 && length != NULL && (*length) >= minLength ) {
         toggled = TRUE;
      }                
      
      if (toggled) {
         tigStore->setUnitigFUR(i, AS_FORCED_UNIQUE);
         numToggled++;
      }      
   }
   
   DeleteHashTable_AS(UIDtoIID);

   delete tigStore;
   
   fprintf(stderr, "Toggled %d\n", numToggled);
   
   return 0;
}
